import 'dart:convert';

import 'package:dio/dio.dart';
import 'package:hive/hive.dart';
import 'package:yuro/yuro_util/yuro_util.dart';

import 'dio_cache_options.dart';

part 'dio_cache.g.dart';

@HiveType(typeId: 201)
class DioCache extends Object {
  static const String REQUEST_SEND_DATE = '@requestSendDate';
  static const String KEY = '@key';
  static const String FROM = '@from';
  static const String MAX_STALE = '@maxStale';
  static const String MAX_AGE = '@maxAge';

  // const _maxAgeHeader = 'max-age';
  // const _maxStaleHeader = 'max-stale';
  // const _minFreshHeader = 'min-fresh';
  // const _mustRevalidateHeader = 'must-revalidate';
  // const _privateHeader = 'private';
  // const _publicHeader = 'public';
  // const _noCacheHeader = 'no-cache';
  // const _noStoreHeader = 'no-store';
  // const _eTagHeader = 'etag';

  /// 缓存key
  @HiveField(12)
  final String originUrl;

  /// 缓存key
  @HiveField(0)
  final String storeKey;

  /// 请求发送的时间
  @HiveField(1)
  DateTime requestDate;

  /// 请求响应的时间
  @HiveField(2)
  DateTime responseDate;

  /// 响应头中的日期
  @HiveField(11)
  DateTime? date;

  /// 最后修改日期
  @HiveField(3)
  DateTime? lastModified;

  /// 最大陈旧时间
  @HiveField(4)
  final DateTime? maxStale;

  /// 最大过期时间
  @HiveField(5)
  final DateTime? expires;

  /// 电子标签
  @HiveField(6)
  final String? eTag;

  /// 响应头
  @HiveField(7)
  final Map<String, dynamic> headers;

  /// 响应内容
  @HiveField(8)
  final List<int>? responseData;

  // /// 缓存控制
  // @HiveField(9)
  // final CacheControl cacheControl;

  DioCache({
    required this.originUrl,
    required this.storeKey,
    required this.requestDate,
    required this.responseDate,
    // required this.cacheControl,
    this.date,
    this.lastModified,
    this.maxStale,
    this.expires,
    this.eTag,
    required this.headers,
    this.responseData,
  });

  factory DioCache.fromResponse(DioCacheOptions options, Response response) {
    Duration? _maxStale = (response.requestOptions.extra[MAX_STALE] as Duration?) ?? options.maxStale;
    Duration? _maxAge = (response.requestOptions.extra[MAX_AGE] as Duration?);
    return DioCache(
        originUrl: response.requestOptions.uri.toString(),
        storeKey: options.keyBuilder(response.requestOptions),
        requestDate: response.requestOptions.extra[REQUEST_SEND_DATE] as DateTime,
        responseDate: DateTime.now(),
        date: response.headers.value('date')?.parseToDateTime(),
        lastModified: response.headers.value('lastModified')?.parseToDateTime(),
        maxStale: _maxStale != null ? DateTime.now().add(_maxStale) : null,
        eTag: response.headers.value('eTag'),
        headers: response.headers.map,
        responseData: _serializeContent(response.requestOptions.responseType, response.data)
        // cacheControl: CacheControl.fromHeaders(response.headers),
        );
  }
}

extension DioCacheExt on DioCache {
  /// 获取DioCache是否已经陈旧
  bool get isStaled => maxStale?.isBefore(DateTime.now()) ?? false;

  /// 获取DioCache是否已经陈旧
  bool get isExpires => expires?.isBefore(DateTime.now()) ?? false;

  Response toResponse(RequestOptions options, bool fromCache) {
    final responseHeaders = Headers();
    headers.forEach((key, value) => responseHeaders.set(key, value));
    return Response(
        requestOptions: options,
        data: _deserializeContent(options.responseType, responseData),
        headers: responseHeaders,
        statusCode: 304,
        extra: {
          DioCache.KEY: storeKey,
          DioCache.FROM: fromCache ? 'cache' : 'server',
        });
  }
}

List<int> _serializeContent(ResponseType type, dynamic content) {
  switch (type) {
    case ResponseType.bytes:
      return content;
    case ResponseType.plain:
      return utf8.encode(content);
    case ResponseType.json:
      return utf8.encode(jsonEncode(content));
    default:
      throw UnsupportedError('Response type not supported : $type.');
  }
}

dynamic _deserializeContent(ResponseType type, List<int>? content) {
  switch (type) {
    case ResponseType.bytes:
      return content;
    case ResponseType.plain:
      return (content != null) ? utf8.decode(content) : null;
    case ResponseType.json:
      return (content != null) ? jsonDecode(utf8.decode(content)) : null;
    default:
      throw UnsupportedError('Response type not supported : $type.');
  }
}
